#include <variable.h>
#include <opengl.h>
#include <vector>
#include <math.h>
#include <iostream>
#include <object.h>

void draw_2p_circle(double xc, double yc, double xe, double ye, double red, double green, double blue, double alpha);
void draw_3p_arc(double xs, double ys, double xc, double yc, double xe, double ye, double red, double green, double blue, double alpha);
void draw_3p_ellipse(double xcen, double ycen, double xend, double yend, double xcon, double ycon, double pi_angle_start, double pi_angle_end, double ratio, double red, double green, double blue, double alpha);
void draw_spline(std::vector<std::vector<double>> C, double red, double green, double blue, double alpha);
void draw_4p_bezier(double xs, double ys, double zs, double xcon0, double ycon0, double zcon0, double xcon1, double ycon1, double zcon1, double xend, double yend, double zend, double red, double green, double blue, double alpha);
void preview_polygon(double xcen, double ycen, double zcen, double xend, double yend, double zend, int sides);

void display(){

    //line preview function
    if(object.type=="line"){
        glLineWidth(1);
        glColor4d(object.red,object.green,object.blue,object.alpha);
        glBegin(GL_LINES);
        glVertex3d(object.start.x,object.start.y,object.start.z);
        glVertex3d(object.end.x,object.end.y,object.end.z);
        glEnd();
    }

    //linestrip preview function
    if(object.type=="linestrip" && object.control.size()>0){
        glLineWidth(1);
        glColor4d(object.red,object.green,object.blue,object.alpha);
        glBegin(GL_LINE_STRIP);
        glVertex3d(object.start.x,object.start.y,object.start.z);
        for(int i=0; i<object.control.size(); i++){
            glVertex3d(object.control.at(i).x,object.control.at(i).y,object.control.at(i).z);
        }
        glVertex3d(object.end.x,object.end.y,object.end.z);
        glEnd();
    }

    //rectangular preview function
    if(object.type=="rectangular"){
        glLineWidth(1);
        glColor4d(object.red,object.green,object.blue,object.alpha);
        glBegin(GL_LINE_LOOP);
        glVertex3d(object.start.x,object.start.y,object.start.z); //lower left
        glVertex3d(object.control.front().x,object.control.front().y,object.control.front().z); //upper left
        glVertex3d(object.end.x,object.end.y,object.end.z); //upper right
        glVertex3d(object.control.back().x,object.control.back().y,object.control.back().z); //lower right
        glEnd();
    }

    //circle preview function
    if(object.type=="circle"){
        draw_2p_circle(object.center.x,object.center.y,object.end.x,object.end.y,object.red,object.green,object.blue,object.alpha);
    }

    //ellipse preview function
    if(object.type=="ellipse"){
        draw_3p_ellipse(object.center.x,object.center.y,object.end.x,object.end.y,object.control.back().x,object.control.back().y,object.pi_angle_start,object.pi_angle_end, object.ratio,object.red,object.green,object.blue,object.alpha);
    }

    //arc preview function
    if(object.type=="arc" && object.control.size()>0){
        draw_3p_arc(object.start.x, object.start.y, object.control.back().x, object.control.back().y, object.end.x, object.end.y, object.red,object.green,object.blue,object.alpha);
    }

    //spline preview function
    if(object.type=="spline" && object.control.size()>0){
        std::vector<std::vector<double>> C; //2d array
        C.push_back({object.start.x, object.start.y}); //spline startpoint
        for(int i=0; i<object.control.size(); i++){
            C.push_back({object.control.at(i).x, object.control.at(i).y}); //spline controlpoints
        }
        C.push_back({object.end.x, object.end.y}); //spline endpoint
        draw_spline(C, object.red,object.green,object.blue,object.alpha);
        C.clear();
    }

    //bezier preview function
    if(object.type=="bezier"){
        draw_4p_bezier(object.start.x,object.start.y,object.start.z,object.control.front().x,object.control.front().y,object.control.front().z,object.control.back().x,object.control.back().y,object.control.back().z,
                       object.end.x,object.end.y,object.end.z, object.red,object.green,object.blue,object.alpha);
    }

    //polygon preview function
    if(object.type=="polygon"){
        preview_polygon(object.center.x,object.center.y,object.center.z,object.end.x,object.end.y,object.end.z,object.sides);
        glLineWidth(1);
        glColor4d(1,1,1,1);
        glBegin(GL_LINE_STRIP);
        glVertex3d(object.start.x,object.start.y,object.start.z);
        for(int i=0; i<object.control.size(); i++){
            glVertex3d(object.control.at(i).x,object.control.at(i).y,object.control.at(i).z);
        }
        glVertex3d(object.end.x,object.end.y,object.end.z);
        glEnd();
    }

    //display stored object data
    for(int i=0; i<objects.size(); i++){
        if(objects.at(i).type=="line"){
            glLineWidth(1);
            glColor4d(objects.at(i).red,objects.at(i).green,objects.at(i).blue,objects.at(i).alpha);
            glBegin(GL_LINES);
            glVertex3d(objects.at(i).start.x,objects.at(i).start.y,objects.at(i).start.z);
            glVertex3d(objects.at(i).end.x,objects.at(i).end.y,objects.at(i).end.z);
            glEnd();
        } else if(objects.at(i).type=="linestrip"){
            glLineWidth(1);
            glColor4d(objects.at(i).red,objects.at(i).green,objects.at(i).blue,objects.at(i).alpha);
            glBegin(GL_LINE_STRIP);
            glVertex3d(objects.at(i).start.x,objects.at(i).start.y,objects.at(i).start.z);
            for(int j=0; j<objects.at(i).control.size(); j++){
                glVertex3d(objects.at(i).control.at(j).x,objects.at(i).control.at(j).y,objects.at(i).control.at(j).z);
            }
            glVertex3d(objects.at(i).end.x,objects.at(i).end.y,objects.at(i).end.z);
            glEnd();
        } else if(objects.at(i).type=="spline"){
            std::vector<std::vector<double>> C; //2d array
            C.push_back({objects.at(i).start.x, objects.at(i).start.y}); //spline startpoint
            for(int j=0; j<objects.at(i).control.size(); j++){
                C.push_back({objects.at(i).control.at(j).x, objects.at(i).control.at(j).y}); //spline controlpoints
            }
            C.push_back({objects.at(i).end.x, objects.at(i).end.y}); //spline endpoint
            draw_spline(C,objects.at(i).red,objects.at(i).green,objects.at(i).blue,objects.at(i).alpha);
            C.clear();
        } else if(objects.at(i).type=="arc"){
            draw_3p_arc(objects.at(i).start.x,objects.at(i).start.y,objects.at(i).control.back().x,objects.at(i).control.back().y,objects.at(i).end.x,objects.at(i).end.y,objects.at(i).red,objects.at(i).green,objects.at(i).blue,objects.at(i).alpha);
        } else if(objects.at(i).type=="circle"){
            draw_2p_circle(objects.at(i).center.x,objects.at(i).center.y,objects.at(i).end.x,objects.at(i).end.y,objects.at(i).red,objects.at(i).green,objects.at(i).blue,objects.at(i).alpha);
        } else if(objects.at(i).type=="ellipse"){
            draw_3p_ellipse(objects.at(i).center.x,objects.at(i).center.y,objects.at(i).end.x,objects.at(i).end.y,objects.at(i).control.back().x,objects.at(i).control.back().y,objects.at(i).pi_angle_start,objects.at(i).pi_angle_end, objects.at(i).ratio,
                            objects.at(i).red,objects.at(i).green,objects.at(i).blue,objects.at(i).alpha);
        } else if(objects.at(i).type=="bezier"){
            draw_4p_bezier(objects.at(i).start.x,objects.at(i).start.y,objects.at(i).start.z,objects.at(i).control.front().x,objects.at(i).control.front().y,objects.at(i).control.front().z,objects.at(i).control.back().x,objects.at(i).control.back().y,
                           objects.at(i).control.back().z,objects.at(i).end.x,objects.at(i).end.y,objects.at(i).end.z,objects.at(i).red,objects.at(i).green,objects.at(i).blue,objects.at(i).alpha);
        } else if(objects.at(i).type=="rectangular"){
            glLineWidth(1);
            glColor4d(objects.at(i).red,objects.at(i).green,objects.at(i).blue,objects.at(i).alpha);
            glBegin(GL_LINE_LOOP);
            glVertex3d(objects.at(i).start.x,objects.at(i).start.y,objects.at(i).start.z); //lower left
            glVertex3d(objects.at(i).control.front().x,objects.at(i).control.front().y,objects.at(i).control.front().z); //upper left
            glVertex3d(objects.at(i).end.x,objects.at(i).end.y,objects.at(i).end.z); //upper right
            glVertex3d(objects.at(i).control.back().x,objects.at(i).control.back().y,objects.at(i).control.back().z); //lower right
            glEnd();
        }
    }
}

void draw_2p_circle(double xc, double yc, double xe, double ye, double red, double green, double blue, double alpha){

    int segments = 25;
    double radius = sqrt(pow(xe-xc,2)+pow(ye-yc,2));
    glColor4d(red,green,blue,alpha);
    glBegin(GL_LINE_LOOP);
    for (int j = 0; j<segments; j++)
    {
        double theta = 2.0 * M_PI * int(j) / int(segments);
        double x = radius * cosf(theta);
        double y = radius * sinf(theta);
        glVertex2d(x + xc, y + yc);
    }
    glEnd();
}

void draw_3p_ellipse(double xcen, double ycen, double xend, double yend, double xcon, double ycon, double pi_angle_start, double pi_angle_end, double ratio, double red, double green, double blue, double alpha){

    //ellipse center = xcen, ycen
    //ellipse base right = xend, yend
    //ellipse top point = xcon, ycon

    double center_x = xcen;
    double center_y = ycen;

    //center point to endpoint mayor axis..
    double radius_x = sqrt(pow(xend-xcen,2) + pow(yend-ycen,2));
    //ratio minor axis to mayor axis..
    double radius_y=0;

    if(ratio==0){ //takes the controlpoint instead of ratio factor
        radius_y = sqrt(pow(xcon-xcen,2) + pow(ycon-ycen,2));
    }
    if(ratio!=0){ //if ratio from dxf is available, it takes its value
        radius_y = ratio*radius_x;
    }

    //calculate angle of ellipse
    double x1=xend;
    double y1=yend;
    double radian = atan2(y1 - center_y, x1 - center_x);
    double angle = radian * (180 / M_PI);
    if (angle < 0.0)
        angle += 360.0;

    //glColor4d(255,255,255,255);
    glColor4d(red,green,blue,alpha);
    glBegin(GL_LINE_STRIP);
    for(double j = pi_angle_start; j<pi_angle_end; j+=0.01)
    {
        //non rotated ellipse..
        double non_rotated_x = /*center_x +*/ cos(j)*radius_x;
        double non_rotated_y = /*center_y +*/ sin(j)*radius_y;
        //glVertex2d(center_x + cos(j)*radius_x, center_y + sin(j)*radius_y);

        //rotated ellipse..
        double rotate_x = center_x  + cos(angle * M_PI / 180.0 )* non_rotated_x - sin(angle * M_PI / 180.0 ) * non_rotated_y;
        double rotate_y = center_y  + sin(angle * M_PI / 180.0 )* non_rotated_x + cos(angle * M_PI / 180.0 ) * non_rotated_y;
        glVertex2d(rotate_x, rotate_y);
    }
    glEnd();
}

void draw_3p_arc(double xs, double ys, double xcon, double ycon, double xe, double ye, double red, double green, double blue, double alpha){

    double x1 = (xcon + xs) / 2;
    double y1 = (ycon + ys) / 2;
    double dy1 = xcon - xs;
    double dx1 = -(ycon - ys);

    // Get the perpendicular bisector of (x2, y2) and (x3, y3).
    double x2 = (xe + xcon) / 2;
    double y2 = (ye + ycon) / 2;
    double dy2 = xe - xcon;
    double dx2 = -(ye - ycon);

    double endpoint_x0 = x1 + dx1;
    double endpoint_y0 = y1 + dy1;
    double endpoint_x1 = x2 + dx2;
    double endpoint_y1 = y2 + dy2;

    //line 1
    double delta_y0 = endpoint_y0 - y1;
    double delta_x0 = x1 - endpoint_x0;
    double c0 = delta_y0 * x1 + delta_x0 * y1;
    //line 2
    double delta_y1 = endpoint_y1 - y2;
    double delta_x1 = x2 - endpoint_x1;
    double c1 = delta_y1 * x2 + delta_x1 * y2;

    double determinant = delta_y0*delta_x1 - delta_y1*delta_x0;
    double deg_start_angle=0;
    double deg_end_angle=0;
    double radius=0;
    double x_center=0;
    double y_center=0;

    if (determinant == 0) // The lines are parallel.
    {
        // lines are parallel
    }
    else  {
        x_center = (delta_x1*c0 - delta_x0*c1)/determinant;
        y_center = (delta_y0*c1 - delta_y1*c0)/determinant;
        double dx = x_center - xs;
        double dy = y_center - ys;
        radius = sqrt(dx * dx + dy * dy);

        //formula if arc is cw or ccw.. this is nice..
        //To determine which side of the line from A=(x1,y1) to B=(x2,y2) a point P=(x,y) falls on you need to compute the value:-
        //d=(x−x1)(y2−y1)−(y−y1)(x2−x1)
        //If d<0 then the point lies on one side of the line, and if d>0 then it lies on the other side. If d=0 then the point lies exactly line.

        double xp = xe;
        double yp = ye;
        double x1 = xs;
        double y1 = ys;
        double x2 = xcon;
        double y2 = ycon;

        double d = ((xp-x1)*(y2-y1))-((yp-y1)*(x2-x1));
        //qDebug() << "d :" << d;

        if(d > 0){
            deg_start_angle = 180/M_PI*atan2(ye-y_center, xe-x_center);    //start angle G2
            deg_end_angle = 180/M_PI*atan2(ys-y_center, xs-x_center);    //end angle G2
        }
        if(d < 0){
            deg_start_angle = 180/M_PI*atan2(ys-y_center, xs-x_center);    //start angle G3
            deg_end_angle  = 180/M_PI*atan2(ye-y_center, xe-x_center);    //end angle G3
        }
        if(d == 0){
            //draw straight line
        }
    }

    //draw arc..
    double pi_angle_start = (deg_start_angle/360)*(2*M_PI); //start angle
    //if(end_angle==360)
    //{
    //    end_angle=0;   //autocad degree problem..
    //}
    double pi_angle_end = (deg_end_angle/360)*(2*M_PI);
    if (pi_angle_end<pi_angle_start)   //this to avoid the start angle is bigger then the end angle value
    {
        pi_angle_end+=2*M_PI;
    }

    double pi_arc_angle = pi_angle_end - pi_angle_start;
    double theta = pi_arc_angle / double(segments - 1);//theta is now calculated from the arc angle instead, the - 1 bit comes from the fact that the arc is open
    double tangetial_factor = tanf(theta);
    double radial_factor = cosf(theta);
    double x = radius * cosf(pi_angle_start);//we now start at the start angle
    double y = radius * sinf(pi_angle_start);

    glColor4d(red,green,blue,alpha);
    glBegin(GL_LINE_STRIP);//since the arc is not a closed curve, this is a strip now
    for(int ii = 0; ii < segments; ii++)
    {
        glVertex2d(x + x_center, y + y_center); //x center, y center..

        float tx = -y;
        float ty = x;

        x += tx * tangetial_factor;
        y += ty * tangetial_factor;

        x *= radial_factor;
        y *= radial_factor;
    }
    glEnd();

}

void draw_spline(std::vector<std::vector<double>> C, double red, double green, double blue, double alpha){
    double S[C.size()-1][2][4];
    double w, b[C.size()], d[C.size()], x[C.size()];
    int i, dim;
    int n = C.size()-1; // number of splines

    for(dim=0; dim<2; dim++)
    {
        // define d[]:
        d[0]=3.0*(C[1][dim]-C[0][dim]);
        for(i=1; i<n; i++)
            d[i]=3.0*(C[i+1][dim]-C[i-1][dim]);
        d[n]=3.0*(C[n][dim]-C[n-1][dim]);
        // forward sweep: (simplified tridiagonal solution: all a[i]=1 and all c[i]=1)
        b[0]=2.0;
        for(i=1; i<C.size(); i++)
        {
            w = 1.0/b[i-1];
            b[i] = 4.0 - w;
            d[i] -= (w*d[i-1]);
        }
        // calculate solution vector x[i] = D[i]:
        // (Wikipedia x[] = Wolfram D[])
        x[n]=d[n]/b[n];
        for(i=n-1; i>=0; i--)
            x[i]=(d[i]-x[i+1])/b[i];
        // calculate spline S[i][dim] a, b, c and d:
        for(i=0; i<n; i++)
        {
            S[i][dim][0]=C[i][dim];
            S[i][dim][1]=x[i];
            S[i][dim][2]=3.0*(C[i+1][dim]-C[i][dim]) - 2.0*x[i] - x[i+1];
            S[i][dim][3]=2.0*(C[i][dim]-C[i+1][dim]) + x[i] + x[i+1];
        }
    }

    glLineWidth(1);
    glColor4d(red,green,blue,alpha);
    glBegin(GL_LINE_STRIP);
    for(int p=0; p<C.size()-1; p++)  //spline points
    {
        for(double t=0; t<1; t+=0.1)  //time 0-1 for each spline
        {
            double px = ((S[p][0][3]*t+S[p][0][2])*t+S[p][0][1])*t+S[p][0][0];
            double py = ((S[p][1][3]*t+S[p][1][2])*t+S[p][1][1])*t+S[p][1][0];
            //std::cout<< "x:" << px << " y:" << py << std::endl;
            glVertex2d(px,py);
        }
    }
    glEnd();

    C.clear(); //empty 2d array
}

void draw_4p_bezier(double xs, double ys, double zs, double xcon0, double ycon0, double zcon0, double xcon1, double ycon1, double zcon1, double xend, double yend, double zend, double red, double green, double blue, double alpha){
    glLineWidth(1);
    glColor4d(red,green,blue,alpha);
    glBegin(GL_LINE_STRIP);
    for(double t = 0; t < 1; t+=0.01){

        //Calculate Point E, on the line AB
        double Ex = ((1-t) * xs) + (t * xcon0);
        double Ey = ((1-t) * ys) + (t * ycon0);
        double Ez = ((1-t) * zs) + (t * zcon0);

        //Calculate Point F, on the line BC
        double Fx = ((1-t) * xcon0) + (t * xcon1);
        double Fy = ((1-t) * ycon0) + (t * ycon1);
        double Fz = ((1-t) * zcon0) + (t * zcon1);

        //Calculate Point G, on the line CD
        double Gx = ((1-t) * xcon1) + (t * xend);
        double Gy = ((1-t) * ycon1) + (t * yend);
        double Gz = ((1-t) * zcon1) + (t * zend);

        //Calculate Point Q, on the line EF
        double Qx = ((1-t) * Ex) + (t * Fx);
        double Qy = ((1-t) * Ey) + (t * Fy);
        double Qz = ((1-t) * Ez) + (t * Fz);

        //Calculate Point R, on the line FG
        double Rx = ((1-t) * Fx) + (t * Gx);
        double Ry = ((1-t) * Fy) + (t * Gy);
        double Rz = ((1-t) * Fz) + (t * Gz);

        //Calculate Point P, on the line QR
        double Px = ((1-t) * Qx) + (t * Rx);
        double Py = ((1-t) * Qy) + (t * Ry);
        double Pz = ((1-t) * Qz) + (t * Rz);

        glVertex3d(Px, Py, Pz);  //draw linestrip..
    }
    glEnd();
}

void preview_polygon(double xcen, double ycen, double zcen, double xend, double yend, double zend, int sides){

    if(sides<3){sides = 3;}
    bool even=0;
    double dist_x = xend - xcen;
    double dist_y = yend - ycen;
    double radius = sqrt(pow(dist_x,2) + pow(dist_y,2));
    double cake_angle = 360 / sides;
    double start_angle = atan2(ycen - yend, xcen - xend)* 180.0 / M_PI;
    double x0,y0,x1,y1;

    if (sides % 2 == 0){
        even = 1;
    } else {
        even = 0;
    }

    object.control.clear();

    if(even == 1){
        for(int i=0; i<sides; i++){
            if(i==0){
                x0 = xcen + (radius * cos ( start_angle * M_PI / 180.0 ));
                y0 = ycen + (radius * sin ( start_angle * M_PI / 180.0 ));
                x1 = xcen + (radius * cos ( (start_angle + cake_angle) * M_PI / 180.0 ));
                y1 = ycen + (radius * sin ( (start_angle + cake_angle) * M_PI / 180.0 ));
                start_angle += cake_angle;
                object.start.x=x0;
                object.start.y=y0;
                object.start.z=zend;
                object.control.push_back({x1,y1,zend});
            }
            if(i>0 && i<sides-1){
                x1 = xcen + (radius * cos ( (start_angle + cake_angle) * M_PI / 180.0 ));
                y1 = ycen + (radius * sin ( (start_angle + cake_angle) * M_PI / 180.0 ));
                start_angle += cake_angle;
                object.control.push_back({x1,y1,zend});
            }
            if(i==sides-1){
                x1 = xcen + (radius * cos ( (start_angle + cake_angle) * M_PI / 180.0 ));
                y1 = ycen + (radius * sin ( (start_angle + cake_angle) * M_PI / 180.0 ));
                object.end.x=x1;
                object.end.y=y1;
                object.end.z=zend;
            }
        }
    }

    if(even == 0){
        for(int i=0; i<sides; i++){
            if(i==0){
                x0 = xcen + (radius * cos ( (start_angle + (0.5*cake_angle)) * M_PI / 180.0 ));
                y0 = ycen + (radius * sin ( (start_angle + (0.5*cake_angle)) * M_PI / 180.0 ));
                x1 = xcen + (radius * cos ( ((start_angle + cake_angle) + (0.5*cake_angle)) * M_PI / 180.0 ));
                y1 = ycen + (radius * sin ( ((start_angle + cake_angle) + (0.5*cake_angle)) * M_PI / 180.0 ));
                start_angle += cake_angle;
                object.start.x=x0;
                object.start.y=y0;
                object.start.z=zend;
                object.control.push_back({x1,y1,zend});
            }
            if(i>0 && i<sides-1){
                x1 = xcen + (radius * cos ( ((start_angle + cake_angle) + (0.5*cake_angle)) * M_PI / 180.0 ));
                y1 = ycen + (radius * sin ( ((start_angle + cake_angle) + (0.5*cake_angle)) * M_PI / 180.0 ));
                start_angle += cake_angle;
                object.control.push_back({x1,y1,zend});
            }
            if(i==sides-1){
                x1 = xcen + (radius * cos ( ((start_angle + cake_angle) + (0.5*cake_angle)) * M_PI / 180.0 ));
                y1 = ycen + (radius * sin ( ((start_angle + cake_angle) + (0.5*cake_angle)) * M_PI / 180.0 ));
                object.end.x=x1;
                object.end.y=y1;
                object.end.z=zend;
            }
        }
    }
}




